@@ -7,7 +7,7 @@ use Config;
 use Text::ParseWords;
 
 use vars qw($VERSION);
-$VERSION = '0.00_01';
+$VERSION = '0.12';
 
 sub new {
   my $class = shift;
@@ -32,7 +32,16 @@ sub find_perl_interpreter {
 
 sub add_to_cleanup {
   my $self = shift;
-  my %files = map {$_, 1} @_;
+  foreach (@_) {
+    $self->{files_to_clean}{$_} = 1;
+  }
+}
+
+sub cleanup {
+  my $self = shift;
+  foreach my $file (keys %{$self->{files_to_clean}}) {
+    unlink $file;
+  }
 }
 
 sub object_file {
@@ -55,6 +64,21 @@ sub arg_object_file {
   return ('-o', $file);
 }
 
+sub arg_share_object_file {
+  my ($self, $file) = @_;
+  return ($self->split_like_shell($self->{config}{lddlflags}), '-o', $file);
+}
+
+sub arg_exec_file {
+  my ($self, $file) = @_;
+  return ('-o', $file);
+}
+
+sub arg_defines {
+  my ($self, %args) = @_;
+  return map "-D$_=$args{$_}", keys %args;
+}
+
 sub compile {
   my ($self, %args) = @_;
   die "Missing 'source' argument to compile()" unless defined $args{source};
@@ -65,13 +89,15 @@ sub compile {
   
   my @include_dirs = $self->arg_include_dirs
     (@{$args{include_dirs} || []},
-     File::Spec->catdir($cf->{installarchlib}, 'CORE'));
+     $self->perl_inc());
+  
+  my @defines = $self->arg_defines( %{$args{defines} || {}} );
   
   my @extra_compiler_flags = $self->split_like_shell($args{extra_compiler_flags});
   my @cccdlflags = $self->split_like_shell($cf->{cccdlflags});
   my @ccflags = $self->split_like_shell($cf->{ccflags});
   my @optimize = $self->split_like_shell($cf->{optimize});
-  my @flags = (@include_dirs, @cccdlflags, @extra_compiler_flags,
+  my @flags = (@include_dirs, @defines, @cccdlflags, @extra_compiler_flags,
 	       $self->arg_nolink,
 	       @ccflags, @optimize,
 	       $self->arg_object_file($args{object_file}),
@@ -128,6 +154,8 @@ sub exe_file {
 
 sub need_prelink { 0 }
 
+sub extra_link_args_after_prelink { return }
+
 sub prelink {
   my ($self, %args) = @_;
   
@@ -139,9 +167,10 @@ sub prelink {
     DL_FUNCS => $args{dl_funcs}     || {},
     FUNCLIST => $args{dl_func_list} || [],
     IMPORTS  => $args{dl_imports}   || {},
-    NAME     => $args{dl_name},
-    DLBASE   => $args{dl_base},
-    FILE     => $args{dl_file},
+    NAME     => $args{dl_name},		# Name of the Perl module
+    DLBASE   => $args{dl_base},		# Basename of DLL file
+    FILE     => $args{dl_file},		# Dir + Basename of symlist file
+    VERSION  => (defined $args{dl_version} ? $args{dl_version} : '0.0'),
   );
   
   # Mksymlists will create one of these files
@@ -157,7 +186,7 @@ sub link_executable {
   my ($self, %args) = @_;
   return $self->_do_link('exe_file', lddl => 0, %args);
 }
-				   
+
 sub _do_link {
   my ($self, $type, %args) = @_;
 
@@ -170,13 +199,17 @@ sub _do_link {
   my @temp_files;
   @temp_files =
     $self->prelink(%args,
-		   dl_name => $args{module_name}) if $self->need_prelink;
+		   dl_name => $args{module_name}) if $args{lddl} && $self->need_prelink;
   
-  my @linker_flags = $self->split_like_shell($args{extra_linker_flags});
-  my @lddlflags = $args{lddl} ? $self->split_like_shell($cf->{lddlflags}) : ();
+  my @linker_flags = ($self->split_like_shell($args{extra_linker_flags}),
+		      $self->extra_link_args_after_prelink(%args, dl_name => $args{module_name},
+							   prelink_res => \@temp_files));
+
+  my @output = $args{lddl} ? $self->arg_share_object_file($out) : $self->arg_exec_file($out);
   my @shrp = $self->split_like_shell($cf->{shrpenv});
   my @ld = $self->split_like_shell($cf->{ld});
-  $self->do_system(@shrp, @ld, @lddlflags, '-o', $out, @$objects, @linker_flags)
+  
+  $self->do_system(@shrp, @ld, @output, @$objects, @linker_flags)
     or die "error building $out from @$objects";
   
   return wantarray ? ($out, @temp_files) : $out;
@@ -185,7 +218,7 @@ sub _do_link {
 
 sub do_system {
   my ($self, @cmd) = @_;
-  print "@cmd\n";
+  print "@cmd\n" if !$self->{quiet};
   return !system(@cmd);
 }
 
@@ -200,4 +233,45 @@ sub split_like_shell {
   return Text::ParseWords::shellwords($string);
 }
 
+# if building perl, perl's main source directory
+sub perl_src {
+  # N.B. makemaker actually searches regardless of PERL_CORE, but
+  # only squawks at not finding it if PERL_CORE is set
+
+  return unless $ENV{PERL_CORE};
+
+  my $Updir = File::Spec->updir;
+  my $dir   = File::Spec->curdir;
+
+  # Try up to 5 levels upwards
+  for (0..5) {
+    if (
+	-f File::Spec->catfile($dir,"config_h.SH")
+	&&
+	-f File::Spec->catfile($dir,"perl.h")
+	&&
+	-f File::Spec->catfile($dir,"lib","Exporter.pm")
+       ) {
+      return $dir;
+    }
+
+    $dir = File::Spec->catdir($dir, $Updir);
+  }
+
+  warn "PERL_CORE is set but I can't find your perl source!\n";
+  return ''; # return empty string if $ENV{PERL_CORE} but can't find dir ???
+}
+
+# directory of perl's include files
+sub perl_inc {
+  my $self = shift;
+
+  $self->perl_src() || File::Spec->catdir($self->{config}{archlibexp},"CORE");
+}
+
+sub DESTROY {
+  my $self = shift;
+  $self->cleanup();
+}
+
 1;
