@@ -8,7 +8,8 @@ use File::Spec;
 
 use ExtUtils::CBuilder::Base;
 
-use vars qw(@ISA);
+use vars qw($VERSION @ISA);
+$VERSION = '0.13';
 @ISA = qw(ExtUtils::CBuilder::Base);
 
 sub new {
@@ -31,6 +32,70 @@ sub _compiler_type {
 	  : 'GCC');
 }
 
+sub split_like_shell {
+  # As it turns out, Windows command-parsing is very different from
+  # Unix command-parsing.  Double-quotes mean different things,
+  # backslashes don't necessarily mean escapes, and so on.  So we
+  # can't use Text::ParseWords::shellwords() to break a command string
+  # into words.  The algorithm below was bashed out by Randy and Ken
+  # (mostly Randy), and there are a lot of regression tests, so we
+  # should feel free to adjust if desired.
+  
+  (my $self, local $_) = @_;
+  
+  return @$_ if defined() && UNIVERSAL::isa($_, 'ARRAY');
+  
+  my @argv;
+  return @argv unless defined() && length();
+  
+  my $arg = '';
+  my( $i, $quote_mode ) = ( 0, 0 );
+  
+  while ( $i < length() ) {
+    
+    my $ch      = substr( $_, $i  , 1 );
+    my $next_ch = substr( $_, $i+1, 1 );
+    
+    if ( $ch eq '\\' && $next_ch eq '"' ) {
+      $arg .= '"';
+      $i++;
+    } elsif ( $ch eq '\\' && $next_ch eq '\\' ) {
+      $arg .= '\\';
+      $i++;
+    } elsif ( $ch eq '"' && $next_ch eq '"' && $quote_mode ) {
+      $quote_mode = !$quote_mode;
+      $arg .= '"';
+      $i++;
+    } elsif ( $ch eq '"' && $next_ch eq '"' && !$quote_mode &&
+	      ( $i + 2 == length()  ||
+		substr( $_, $i + 2, 1 ) eq ' ' )
+	    ) { # for cases like: a"" => [ 'a' ]
+      push( @argv, $arg );
+      $arg = '';
+      $i += 2;
+    } elsif ( $ch eq '"' ) {
+      $quote_mode = !$quote_mode;
+    } elsif ( $ch eq ' ' && !$quote_mode ) {
+      push( @argv, $arg ) if $arg;
+      $arg = '';
+      ++$i while substr( $_, $i + 1, 1 ) eq ' ';
+    } else {
+      $arg .= $ch;
+    }
+    
+    $i++;
+  }
+  
+  push( @argv, $arg ) if defined( $arg ) && length( $arg );
+  return @argv;
+}
+
+sub arg_defines {
+  my ($self, %args) = @_;
+  s/"/\\"/g foreach values %args;
+  return map qq{"-D$_=$args{$_}"}, keys %args;
+}
+
 sub compile {
   my ($self, %args) = @_;
   my $cf = $self->{config};
@@ -40,6 +105,10 @@ sub compile {
   my ($basename, $srcdir) =
     ( File::Basename::fileparse($args{source}, '\.[^.]+$') )[0,1];
 
+  $srcdir ||= File::Spec->curdir();
+
+  my @defines = $self->arg_defines( %{ $args{defines} || {} } );
+
   my %spec = (
     srcdir      => $srcdir,
     builddir    => $srcdir,
@@ -50,19 +119,18 @@ sub compile {
     cflags      => [
                      $self->split_like_shell($cf->{ccflags}),
                      $self->split_like_shell($cf->{cccdlflags}),
+                     $self->split_like_shell($cf->{extra_compiler_flags}),
                    ],
     optimize    => [ $self->split_like_shell($cf->{optimize})    ],
-    defines     => [ '' ],
-    includes    => $self->{include_dirs} || [],
+    defines     => \@defines,
+    includes    => [ @{$args{include_dirs} || []} ],
     perlinc     => [
-                     File::Spec->catdir($cf->{archlib}, 'CORE'),
+                     $self->perl_inc(),
                      $self->split_like_shell($cf->{incpath}),
                    ],
     use_scripts => 1, # XXX provide user option to change this???
   );
 
-  $self->add_to_cleanup($spec{output});
-
   $self->normalize_filespecs(
     \$spec{source},
     \$spec{output},
@@ -88,21 +156,29 @@ sub link {
 
   my @objects = ( ref $args{objects} eq 'ARRAY' ? @{$args{objects}} : $args{objects} );
   my $to = join '', (File::Spec->splitpath($objects[0]))[0,1];
+  $to ||= File::Spec->curdir();
 
   (my $file_base = $args{module_name}) =~ s/.*:://;
+  my $output = $args{lib_file} ||
+    File::Spec->catfile($to, "$file_base.$cf->{dlext}");
+
+  # if running in perl source tree, look for libs there, not installed
+  my $lddlflags = $cf->{lddlflags};
+  my $perl_src = $self->perl_src();
+  $lddlflags =~ s/\Q$cf->{archlibexp}\E[\\\/]CORE/$perl_src/ if $perl_src;
 
   my %spec = (
     srcdir        => $to,
     builddir      => $to,
     startup       => [ ],
-    objects       => @objects,
+    objects       => \@objects,
     libs          => [ ],
-    output        => File::Spec->catfile($to, "$file_base.$cf->{dlext}"),
+    output        => $output,
     ld            => $cf->{ld},
     libperl       => $cf->{libperl},
     perllibs      => [ $self->split_like_shell($cf->{perllibs})  ],
     libpath       => [ $self->split_like_shell($cf->{libpth})    ],
-    lddlflags     => [ $self->split_like_shell($cf->{lddlflags}) ],
+    lddlflags     => [ $self->split_like_shell($lddlflags) ],
     other_ldflags => [ $self->split_like_shell($args{extra_linker_flags} || '') ],
     use_scripts   => 1, # XXX provide user option to change this???
   );
@@ -116,6 +192,8 @@ sub link {
 
   $spec{output}    ||= File::Spec->catfile( $spec{builddir},
                                             $spec{basename}  . '.'.$cf->{dlext}   );
+  $spec{manifest}  ||= File::Spec->catfile( $spec{builddir},
+                                            $spec{basename}  . '.'.$cf->{dlext}.'.manifest');
   $spec{implib}    ||= File::Spec->catfile( $spec{builddir},
                                             $spec{basename}  . $cf->{lib_ext} );
   $spec{explib}    ||= File::Spec->catfile( $spec{builddir},
@@ -127,10 +205,10 @@ sub link {
 
   $self->add_to_cleanup(
     grep defined,
-    @{[ @spec{qw(output implib explib def_file base_file map_file)} ]}
+    @{[ @spec{qw(manifest implib explib def_file base_file map_file)} ]}
   );
 
-  foreach my $opt ( qw(output implib explib def_file map_file base_file) ) {
+  foreach my $opt ( qw(output manifest implib explib def_file map_file base_file) ) {
     $self->normalize_filespecs( \$spec{$opt} );
   }
 
@@ -149,9 +227,9 @@ sub link {
     $self->do_system( @$cmd );
   }
 
-#  (my $out = $spec{output}) =~ tr/'"//d;
+  $spec{output} =~ tr/'"//d;
   return wantarray
-    ? grep defined, @spec{qw[output implib explib def_file map_file base_file]}
+    ? grep defined, @spec{qw[output manifest implib explib def_file map_file base_file]}
     : $spec{output};
 }
 
@@ -173,6 +251,19 @@ sub normalize_filespecs {
   }
 }
 
+# directory of perl's include files
+sub perl_inc {
+  my $self = shift;
+
+  my $perl_src = $self->perl_src();
+
+  if ($perl_src) {
+    File::Spec->catdir($perl_src, "lib", "CORE");
+  } else {
+    File::Spec->catdir($self->{config}{archlibexp},"CORE");
+  }
+}
+
 1;
 
 ########################################################################
@@ -180,10 +271,10 @@ sub normalize_filespecs {
 =begin comment
 
 The packages below implement functions for generating properly
-formated commandlines for the compiler being used. Each package
+formatted commandlines for the compiler being used. Each package
 defines two primary functions 'format_linker_cmd()' &
 'format_compiler_cmd()' that accepts a list of named arguments (a
-hash) and returns a list of formated options suitable for invoking the
+hash) and returns a list of formatted options suitable for invoking the
 compiler. By default, if the compiler supports scripting of its
 operation then a script file is built containing the options while
 those options are removed from the commandline, and a reference to the
@@ -210,7 +301,7 @@ sub format_compiler_cmd {
     if $spec{use_scripts};
 
   return [ grep {defined && length} (
-    $spec{cc}, '-c'         ,
+    $spec{cc},'-nologo','-c',
     @{$spec{includes}}      ,
     @{$spec{cflags}}        ,
     @{$spec{optimize}}      ,
@@ -228,8 +319,7 @@ sub write_compiler_script {
                                     $spec{basename} . '.ccs' );
 
   $self->add_to_cleanup($script);
-
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
@@ -250,20 +340,26 @@ sub write_compiler_script {
 
 sub format_linker_cmd {
   my ($self, %spec) = @_;
+  my $cf = $self->{config};
 
   foreach my $path ( @{$spec{libpath}} ) {
     $path = "-libpath:$path";
   }
 
-  $spec{def_file}  &&= '-def:'    . $spec{def_file};
-  $spec{output}    &&= '-out:'    . $spec{output};
-  $spec{implib}    &&= '-implib:' . $spec{implib};
-  $spec{map_file}  &&= '-map:'    . $spec{map_file};
+  my $output = $spec{output};
+
+  $spec{def_file}  &&= '-def:'      . $spec{def_file};
+  $spec{output}    &&= '-out:'      . $spec{output};
+  $spec{manifest}  &&= '-manifest ' . $spec{manifest};
+  $spec{implib}    &&= '-implib:'   . $spec{implib};
+  $spec{map_file}  &&= '-map:'      . $spec{map_file};
 
   %spec = $self->write_linker_script(%spec)
     if $spec{use_scripts};
 
-  return [ grep {defined && length} (
+  my @cmds; # Stores the series of commands needed to build the module.
+
+  push @cmds, [ grep {defined && length} (
     $spec{ld}               ,
     @{$spec{lddlflags}}     ,
     @{$spec{libpath}}       ,
@@ -277,6 +373,15 @@ sub format_linker_cmd {
     $spec{implib}           ,
     $spec{output}           ,
   ) ];
+
+  if ($cf->{cc} eq 'cl' and $cf->{ccversion} =~ /^(\d+)/ and $1 >= 14) {
+    # Embed the manifest file for VC 2005 (aka VC 8) or higher
+    push @cmds, [
+      'mt', '-nologo', $spec{manifest}, '-outputresource:' . "$output;2"
+    ];
+  }
+
+  return @cmds;
 }
 
 sub write_linker_script {
@@ -287,7 +392,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
@@ -344,16 +449,21 @@ sub write_compiler_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
 
+  # XXX Borland "response files" seem to be unable to accept macro
+  # definitions containing quoted strings. Escaping strings with
+  # backslash doesn't work, and any level of quotes are stripped. The
+  # result is is a floating point number in the source file where a
+  # string is expected. So we leave the macros on the command line.
   print SCRIPT join( "\n",
     map { ref $_ ? @{$_} : $_ }
     grep defined,
     delete(
-      @spec{ qw(includes cflags optimize defines perlinc) } )
+      @spec{ qw(includes cflags optimize perlinc) } )
   );
 
   close SCRIPT;
@@ -404,7 +514,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($ld_script, $ld_libs);
 
-  print "Generating scripts '$ld_script' and '$ld_libs'.\n";
+  print "Generating scripts '$ld_script' and '$ld_libs'.\n" if !$self->{quiet};
 
   # Script 1: contains options & names of object files.
   open( LD_SCRIPT, ">$ld_script" )
@@ -449,8 +559,11 @@ sub format_compiler_cmd {
     $path = '-I' . $path;
   }
 
+  # split off any -arguments included in cc
+  my @cc = split / (?=-)/, $spec{cc};
+
   return [ grep {defined && length} (
-    $spec{cc}, '-c'         ,
+    @cc, '-c'               ,
     @{$spec{includes}}      ,
     @{$spec{cflags}}        ,
     @{$spec{optimize}}      ,
@@ -495,8 +608,11 @@ sub format_linker_cmd {
                '--output-exp' , $spec{explib}
   ];
 
+  # split off any -arguments included in ld
+  my @ld = split / (?=-)/, $spec{ld};
+
   push @cmds, [ grep {defined && length} (
-    $spec{ld}                 ,
+    @ld                       ,
     '-o', $spec{output}       ,
     "-Wl,--base-file,$spec{base_file}"   ,
     "-Wl,--image-base,$spec{image_base}" ,
@@ -518,7 +634,7 @@ sub format_linker_cmd {
   ];
 
   push @cmds, [ grep {defined && length} (
-    $spec{ld}                 ,
+    @ld                       ,
     '-o', $spec{output}       ,
     "-Wl,--image-base,$spec{image_base}" ,
     @{$spec{lddlflags}}       ,
@@ -543,7 +659,7 @@ sub write_linker_script {
 
   $self->add_to_cleanup($script);
 
-  print "Generating script '$script'\n";
+  print "Generating script '$script'\n" if !$self->{quiet};
 
   open( SCRIPT, ">$script" )
     or die( "Could not create script '$script': $!" );
